from .angular_loss import AngularLoss, AngularNPairsLoss
from .contrastive_loss import ContrastiveLoss
from .generic_pair_loss import GenericPairLoss
from .lifted_structure_loss import GeneralizedLiftedStructureLoss
from .margin_loss import MarginLoss
from .multi_similarity_loss import MultiSimilarityLoss
from .nca_loss import NCALoss
from .n_pairs_loss import NPairsLoss
from .proxy_losses import ProxyNCALoss
from .triplet_margin_loss import TripletMarginLoss